﻿using Carmenta.Engine;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;

namespace CustomDataSetCSharp
{
    /// <summary>
    /// Custom data set that holds point features read from a text file with comma separated values.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This custom data set caches modifiable features between updates which means that it cannot
    /// implement <b>IMinimizeGarbageCollection</b>.
    /// </para>
    /// </remarks>
    public class CsvDataSet : CustomDataSetAdapter
    {
        /// <summary>
        /// Dictionary that maps ids to features.
        /// </summary>
        private readonly Dictionary<ulong, Feature> features = new Dictionary<ulong, Feature>();

        /// <summary>
        /// Bounding rectangle for the data set.
        /// </summary>
        private Rectangle bounds = Rectangle.Empty;

        /// <summary>
        /// The path to the directory containing the file from which the features in this data set were loaded.
        /// </summary>
        private string path = "";

        /// <summary>
        /// The name of the file from which the features in this data set were loaded.
        /// </summary>
        private string fileName = "";

        /// <summary>
        /// Initializes a new instance of the <see cref="CsvDataSet"/> class.
        /// </summary>
        public CsvDataSet() : base(
                Crs.Wgs84LongLat,
                false)   // Specifies that this data set caches modifiable features between updates
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CsvDataSet"/> class that reads from the CSV file at the given
        /// path and fileName.
        /// </summary>
        public CsvDataSet(string path, string fileName) : base(
                Crs.Wgs84LongLat,
                false)   // Specifies that this data set caches modifiable features between updates
        {
            this.path = path;
            this.fileName = fileName;
        }

        /// <summary>
        /// Overridden, returns a copy of the <b>CsvDataSet</b>.
        /// </summary>
        /// <returns>A copy of the data set.</returns>
        public override ICustomDataSet OnClone()
        {
            return new CsvDataSet();
        }

        /// <summary>
        /// Overridden, initializes the data set with point features from a text file.
        /// </summary>
        /// <param name="context">
        /// Context object that provides access to the <b>CustomDataSetProxy</b>.
        /// </param>
        public override void OnInitialize(CustomDataSetContext context)
        {
            // If this CsvDataSet did not get a file path on construction, try to get it from CustomDataSetProxy.userProperties.
            if (fileName.Length == 0)
            {
                if (context.UserProperties.TryGetValue("input", out AttributeValue file) && file.Type == AttributeType.String)
                {
                    // Loads the text file specified by the fileName attribute in CustomDataSetProxy.userProperties
                    fileName = (string)file.Value;
                }
                else
                {
                    // Loads point.txt if another file has not been specified in CustomDataSetProxy.userProperties
                    fileName = "data.txt";
                }

                if (path.Length == 0)
                {
                    path = context.ConfigurationPath;
                }
            }

            string fullPath = path.Length > 0 ? path + Path.DirectorySeparatorChar + fileName : fileName;
            Load(fullPath);
        }

        /// <summary>
        /// Overridden, returns the <b>Feature</b> with the specified id. This override is necessary if we
        /// want to be able to select features from this data set.
        /// </summary>
        /// <param name="objectId">The id of the <b>Feature</b> to get.</param>
        /// <returns>The <b>Feature</b> with the specified id if found; otherwise, <b>null</b>.</returns>
        public override Feature OnGetFeature(ulong objectId)
        {
            return features.TryGetValue(objectId, out Feature feature) ? feature : null;
        }

        /// <summary>
        /// Overridden, returns all features inside the specified area that match the specified query.
        /// </summary>
        /// <param name="area">The rectangle to return features from.</param>
        /// <param name="query">A query which should match all returned features.</param>
        /// <param name="info">Information about the update operation.</param>
        /// <returns>
        /// A <b>CustomDataSetResult</b> that contains all features inside <i>area</i> that match
        /// <i>query</i>. The <i>info</i> parameter is not used in this sample.
        /// </returns>
        public override CustomDataSetResult OnQuery(Rectangle area, Query query, ViewInfo info)
        {
            // Create a result object
            CustomDataSetResult result = new CustomDataSetResult();

            // Iterate through all features in the data set
            foreach (KeyValuePair<ulong, Feature> obj in features)
            {
                Feature feature = obj.Value;

                // Skip this feature if its bounds does not overlap the specified area
                if (!feature.CalculateBounds().Overlap(area))
                {
                    continue;
                }

                // Skip this feature if a query has been specified that does not match the feature
                if (query != null && !query.Match(feature))
                {
                    continue;
                }

                // The feature is inside the specified area and matches the query so return it
                result.Add(feature);
            }

            // Return the result
            return result;
        }

        /// <summary>
        /// Returns the bounding rectangle spanning all features in this data set.
        /// </summary>
        public override Rectangle OnCalculateBounds()
        {
            return bounds;
        }

        /// <summary>
        /// Loads a point features from a text file with comma separated values. Each line represents
        /// a single feature.
        /// </summary>
        /// <param name="fileName">The file name of a text file with comma separated values</param>
        private void Load(string fileName)
        {
            foreach (string line in File.ReadAllLines(fileName))
            {
                LoadFeature(line);
            }
        }

        /// <summary>
        /// Loads a single point feature from a <b>string</b> with comma separated values.
        /// </summary>
        /// <param name="s">
        /// A <b>string</b> with the following format: "Long,Lat,Name,Population".
        /// </param>
        /// <remarks>
        /// <para>
        /// We observer our data input to recognize the pattern: lat, lon, name, population. Now we shall parse
        /// this information and generate point features for each line, in total 4 lines.
        /// To do this, we create a point feature and two attributes, NAME and POPULATION.
        /// </para>
        /// </remarks>
        private void LoadFeature(string s)
        {
            // Split the string into separate values
            string[] strs = s.Split(',');

            if (strs.Length == 4)
            {
                // Parse the longitude and latitude, skip this entry if we fail to parse the coordinates
                if (double.TryParse(strs[1], NumberStyles.Any, CultureInfo.InvariantCulture.NumberFormat, out double x) &&
                    double.TryParse(strs[0], NumberStyles.Any, CultureInfo.InvariantCulture.NumberFormat, out double y))
                {
                    // Create a unique id for the new feature
                    ulong objectId = (ulong)features.Count;

                    // Create the feature at the specified coordinates
                    Feature feature = new Feature(
                        new PointGeometry(x, y),
                        Crs.Wgs84LongLat,
                        new Id(0, objectId));

                    // Set the name & population of the feature
                    feature.Attributes["NAME"] = strs[2];
                    feature.Attributes["POPULATION"] = double.Parse(strs[3]);

                    // Save the feature in the data set
                    features[objectId] = feature;

                    // Update the bounding rectangle that contains all features in the data set
                    bounds = features.Count == 1
                        ? feature.CalculateBounds()
                        : new Rectangle(
                            Math.Min(bounds.XMin, x),
                            Math.Min(bounds.YMin, y),
                            Math.Max(bounds.XMax, x),
                            Math.Max(bounds.YMax, y));
                }
            }
        }
    }
}
